
(require-extension posix)

;;file-exists?
(define file-regular? regular-file?)
(define file-directory? directory?)
(define file-readable? file-read-access?)
(define file-writable? file-write-access?)
;;file-modification-time
;;file-size
;;create-directory
;;delete-file
;;rename-file

(define (copy-file from to)
  (process-run "cp" from to))

(define (create-directory* f)
  (let* ((end (- (string-length f) 1))
         (f/ (if (eqv? #\/ (string-ref f end)) (substring f 0 end) f)))
    (or (file-exists? f/)
        (condition-case (create-directory f/)
         (e (exn file)
           (let ((dir (pathname-directory f/)))
             (and (string? dir)
                  (not (string=? dir f/))
                  (create-directory* dir)
                  (create-directory f/))))))))

(define (directory-fold dir kons knil)
  (let lp ((ls (cddr (directory dir))) (acc knil))
    (if (null? ls)
      acc
      (lp (cdr ls) (kons (car ls) acc)))))

(define (directory-fold-tree dir kons knil)
  (let lp ((ls (directory-files dir))
           (d dir)
           (hl (- (vector-ref (file-stat dir) 2) 2))
           (acc knil))
    (cond
      ((null? ls)
       acc)
      ((and-let* (((positive? hl))
                  (d2 (string-append d "/" (car ls)))
                  ((file-directory? d2)))
         d2)
       =>
       (lambda (d2)
         (lp (cdr ls)
             d
             (- hl 1)
             (kons d (car ls)
                   (lp (directory-files d2)
                       d2
                       (- (vector-ref (file-stat dir) 2) 1)
                       acc)))))
      (else
       (lp (cdr ls) d hl (kons d (car ls) acc))))))

(define (directory-files dir)
  (cddr (directory dir)))

(define (find-in-path file path . o)
  (let ((pred (if (string? file) (lambda (f) (string=? f file)) file))
        (ls (if (string? path) (list path) path)))
    (let-optionals* o ((guard identity)
                       (recurse? #f))
      (cond
        (recurse?
         (call-with-current-continuation
          (lambda (found)
            (for-each
             (lambda (dir)
               (directory-fold-tree
                dir
                (lambda (d f a)
                  (let ((file (string-append d "/" f)))
                    (cond ((and (pred file) (guard file)) => found))))
                #f))
             ls))))
        ((not (string? file))
         (call-with-current-continuation
          (lambda (found)
            (for-each
             (lambda (dir)
               (directory-fold
                dir
                (lambda (f a)
                  (cond ((and (pred file) (guard file)) => found)))
                #f))
             ls))))
        (else
         (let lp ((ls ls))
           (and (pair? ls)
                (let* ((dir (car ls))
                       (len (string-length dir))
                       (path (if (eqv? #\/ (string-ref dir (- len 1)))
                               (string-append dir file)
                               (string-append dir "/" file))))
                  (or (and (file-exists? path) (guard path))
                      (lp (cdr ls)))))))))))

