
(require-extension posix extras)

(cond-expand
 ((and chicken compiling)
  (declare
   (export process-run process-run-and-exit
           process-output->string process-output->string-list
           with-input-from-process with-output-to-process
           with-process-io shell-escape)))
 (else))

(define (shell-escape-char? c)
  (not (or (char-alphabetic? c)
           (char-numeric? c)
           (memv c '(#\_ #\- #\+ #\: #\, #\= #\/ #\.)))))

(define (shell-escape-backslash str)
  (with-output-to-string
    (lambda ()
      (with-input-from-string str
        (lambda ()
          (let lp ()
            (let ((c (read-char)))
              (if (not (eof-object? c))
                (begin
                  (if (shell-escape-char? c)
                    (display #\\))
                  (display c)
                  (lp))))))))))

(define (shell-escape str)
  (let lp ((i (- (string-length str) 1)) (quote? #f))
    (if (negative? i)
      (if quote?
        (string-append "\"" str "\"")
        str)
      (let ((c (string-ref str i)))
        (cond
          ((or (eqv? c #\") (eqv? c #\\))
           (shell-escape-backslash str))
          ((shell-escape-char? c)
           (lp (- i 1) #t))
          (else
           (lp (- i 1) quote?)))))))

(define (make-process-string command ls)
  (string-intersperse
   (cons command (map shell-escape (map ->string ls)))))

(define (process-run command . args)
  (zero? (system (make-process-string command args))))

(define (process-run-and-exit command . args)
  (process-execute command (map ->string args)))

(define (with-input-from-process x thunk)
  (with-input-from-pipe
   (if (pair? x) (make-process-string (car x) (cdr x)) x)
   thunk))

(define (with-output-to-process x thunk)
  (with-output-to-pipe
   (if (pair? x) (make-process-string (car x) (cdr x)) x)
   thunk))

(define (with-process-io x thunk)
  (let ((command (if (pair? x) (make-process-string (car x) (cdr x)) x)))
    (receive (in out id) (process command)
      (with-input-from-port in
        (lambda ()
          (with-output-to-port out
            (lambda ()
              (let ((res (thunk)))
                (close-input-port in)
                (close-output-port out)
                res))))))))

(define (process-output->string-list x . args)
  (let ((command (make-process-string x args)))
    (with-input-from-process command
      (lambda ()
        (let lp ((res '()))
          (let ((line (read-line)))
            (if (eof-object? line)
              (reverse res)
              (lp (cons line res)))))))))

(define (process-output->string . o)
  (string-intersperse (apply process-output->string-list o) "\n"))
