
(declare
 (export sort sort! sort3 sort3! string-compare3 compare3 object<?))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some SRFI-43 utils

(define (vector-copy v . o)
  (let-optionals* o ((start 0)
                     (end (vector-length v))
                     (fill #f))
    (let* ((len (- end start))
           (res (make-vector len fill))
           (limit (min end (vector-length v))))
      (do ((i start (+ i 1))
           (j 0 (+ j 1)))
          ((= i limit) res)
        (vector-set! res j (vector-ref v i))))))

(define (vector-copy! target tstart source . o)
  (let-optionals* o ((start 0)
                     (end (vector-length source)))
    (if (> start tstart)
      (do ((i tstart (+ i 1))
           (j start (+ j 1)))
          ((= j end) target)
        (vector-set! target i (vector-ref source j)))
      (do ((i (+ tstart (- end start) -1) (- i 1))
           (j (- end 1) (- j 1)))
          ((< j start) target)
        (vector-set! target i (vector-ref source j))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; general scheme object ordering

(define (string-compare3 a b)
  (let* ((len-a (string-length a))
         (len-b (string-length b))
         (len (min len-a len-b)))
    (let lp ((i 0))
      (if (= i len)
        (- len-a len-b)
        (let ((diff (- (char->integer (string-ref a i))
                       (char->integer (string-ref b i)))))
          (if (zero? diff)
            (lp (+ i 1))
            diff))))))

(define (compare3 a b)
  (cond
    ((number? a)
     (if (number? b) (- a b) -1))
    ((char? a)
     (cond
       ((number? b) 1)
       ((char? b) (- (char->integer a) (char->integer b)))
       (else -1)))
    ((string? a)
     (cond
       ((or (char? b) (number? b)) 1)
       ((string? b) (string-compare3 a b))
       (else -1)))
    ((symbol? a)
     (cond
       ((or (string? b) (char? b) (number? b)) 1)
       ((symbol? b) (string-compare3 (symbol->string a) (symbol->string b)))
       (else -1)))
    ((boolean? a)
     (cond
       ((or (string? b) (char? b) (number? b)) 1)
       ((boolean? b) (if a (if b 0 -1) (if b 1 0)))
       (else -1)))
    (else
     (if (or (number? b) (char? b) (string? b) (symbol? b) (boolean? b))
       -1
       0))))

(define (object<? a b)
  (negative? (compare3 a b)))

(define (sort! seq . o)
  (if (vector? seq)
    (apply vector-sort! seq o)
    (vector->list (apply vector-sort! (list->vector seq) o))))

(define (sort seq . o)
  (if (vector? seq)
    (apply vector-sort! (vector-copy seq) o)
    (vector->list (apply vector-sort! (list->vector seq) o))))

(define (vector-sort! v . o)
  (let-optionals* o ((less object<?)
                     (key identity)
                     (tmp #f))
    (if (eq? key identity)
      (let* ((len (vector-length v))
             (v2 (or tmp (make-vector len))))
        (let lp ((lo 0) (hi (- len 1)))
          (let ((diff (- hi lo)))
            (if (< diff 2)
              (if (= diff 1)
                (let ((a (vector-ref v lo))
                      (b (vector-ref v hi)))
                  (if (less b a)
                    (begin
                      (vector-set! v lo b)
                      (vector-set! v hi a)))))
              (let ((mid (+ lo (quotient diff 2))))
                ;; sort subvectors of v
                (lp lo mid)
                (lp (+ mid 1) hi)
                ;; merge into v2
                (let lp2 ((i lo) (j (+ mid 1)) (k lo))
                  (cond
                    ;; end case, copy back into v
                    ((or (> i mid) (> j hi))
                     (let ((off (- k i)))
                       (if (negative? off)
                         (do ((i i (+ i 1))) ((> i mid))
                           (vector-set! v (+ i off) (vector-ref v i)))
                         (do ((j mid (- j 1))) ((< j i))
                           (vector-set! v (+ j off) (vector-ref v j)))))
                     (do ((j lo (+ j 1))) ((= j k))
                       (vector-set! v j (vector-ref v2 j)))
;                      (vector-copy! v k v i mid)
;                      (vector-copy! v lo v2 lo k)
                     )
                    ;; compare and copy one element into v2
                    (else
                     (let ((a (vector-ref v i))
                           (b (vector-ref v j)))
                       (cond
                         ((not (less b a))
                          (vector-set! v2 k a)
                          (lp2 (+ i 1) j (+ k 1)))
                         (else
                          (vector-set! v2 k b)
                          (lp2 i (+ j 1) (+ k 1))))))))))))
        v)
      (list->vector
       (map cdr
            (sort
             (map (lambda (x) (cons (key x) x)) (vector->list v))
             (lambda (a b) (less (car a) (car b)))))))))

(define (sort3! seq . o)
  (if (vector? seq)
    (apply vector-sort3! seq o)
    (vector->list (apply vector-sort3! (list->vector seq) o))))

(define (sort3 seq . o)
  (if (vector? seq)
    (apply vector-sort3! (vector-copy seq) o)
    (vector->list (apply vector-sort3! (list->vector seq) o))))

(define (x->comparator x)
  (cond ((procedure? x) x)
        ((pair? x) (x->comparator (list->vector x)))
        (else
         (let ((len (vector-length x)))
           (lambda (a b)
             (let lp ((i 0))
               (if (= i len)
                 0
                 (let ((diff ((vector-ref x i) a b)))
                   (if (zero? diff)
                     (lp (+ i 1))
                     diff)))))))))

(define (vector-sort3! v . o)
  (let-optionals* o ((cmp compare3)
                     (key identity))
    (cond
      ((pair? key)
       (vector-sort3! v cmp (list->vector key)))
      ((vector? key)
       (let* ((len (vector-length v))
              (tmp (make-vector len))
              (key-len (vector-length key))
              (cmps (if (vector? cmp) cmp (make-vector key-len cmp))))
         ;; build a less-than comparator which lazily computes and
         ;; compares successive keys.
         (define (less v1 v2)
           (let lp ((i 0))
             (if (= i key-len)
               0
               (let ((a (vector-ref v1 i))
                     (b (vector-ref v2 i)))
                 (if (not a)
                   (let ((get-key (vector-ref key i)))
                     (set! a (get-key (vector-ref v1 key-len)))
                     (vector-set! v1 i a)))
                 (if (not b)
                   (let ((get-key (vector-ref key i)))
                     (set! b (get-key (vector-ref v2 key-len)))
                     (vector-set! v2 i b)))
                 (let ((diff ((vector-ref cmps i) a b)))
                   (if (zero? diff)
                     (lp (+ i 1))
                     (negative? diff)))))))
         ;; build a new vector which maps each original element into a
         ;; vector containing that element plus room for all key
         ;; computations.
         (do ((i 0 (+ i 1))) ((= i len))
           (let ((key-vec (make-vector (+ key-len 1) #f)))
             (vector-set! key-vec key-len (vector-ref v i))
             (vector-set! tmp i key-vec)))
         ;; sort it
         (vector-sort! tmp less identity v)
         ;; copy the results back into v
         (do ((i 0 (+ i 1))) ((= i len) v)
           (vector-set! v i (vector-ref (vector-ref tmp i) key-len)))))
       (else
        (let ((cmp (x->comparator cmp)))
          (vector-sort! v (lambda (a b) (negative? (cmp a b))) key))))))
