
(require-extension syntax-case)

(define-macro (common-module name clauses . body)
  (let ((res (generate-common-module name clauses body)))
    ;;(fprintf (current-error-port) "common-module: ~S\n" res)
    res))

(define-macro (import-extension . args)
  (let ((mods (expand-module-symbol-list args)))
    (let lp ((ls mods) (imps '()))
      (cond
        ((null? ls)
         `(begin
            ,(if (pair? mods) `(require-extension ,@mods) ())
            ,@(map (lambda (m) `(import ,m)) imps)))
        ((psyntax-module? (car ls))
         (lp (cdr ls) (cons (car ls) imps)))
        (else
         (lp (cdr ls) imps))))))

(define-macro (import-syntax . args)
  `(import-extension ,@args))

(define-macro (import-lazy module . procs)
  `(import-extension ,module))

(define-macro (let-keywords* ls params . body)
  `(apply (lambda (#!key ,@params) ,@body) ,ls))

(define srfi-aliases
  '((cond-expand . 0)
    (list-lib . 1)
    (and-let* . 2)
    (numeric-vectors . 4)
    (signature-let . 5)
    (string-ports . 6)
    (configuration-language . 7)
    (receive . 8)
    (define-record-type . 9)
    (external-form-syntax . 10)
    (multiple-value-bind . 11)
    (string-lib . 13)
    (char-set-lib . 14)
    (case-lambda . 16)
    (generalized-set! . 17)
    (multithreading . 18)
    (time-lib . 19)
    (real-time-multithreading . 21)
    (unix-scheme-scripts . 22)
    (error . 23)
    (array-lib . 25)
    (cut . 26)
    (random-numbers . 27)
    (basic-format . 28)
    (localization . 29)
    (multi-line-comment-syntax . 30)
    (rec . 31)
    (sort-lib . 32)
    (bitwise-lib . 33)
    (exceptions . 34)
    (conditions . 35)
    (i/o-conditions . 36)
    (args-fold . 37)
    (shared-structure-read/write . 38)
    (parameters . 39)
    (streams-lib . 40)
    (comprehensions . 42)
    (vector-lib . 43)
    (comprehensions . 44)
    (lazy-eval . 45)
    (syntax-rules-choose-ellipse . 46)
    (homogeneous-array-lib . 47)
    (intermediate-format . 48)
    (require-extension . 55)
    (advanced-records . 57)
    (array-notation . 58)
    (vicinity . 59)
    (bitwise-operations . 60)
    (sexpr-comments . 62)
    (homogeneous-and-heterogeneous-arrays . 63)
    (testing . 64)
    ))

(define (append-map1 f ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (append (reverse (f (car ls))) res)))))

(define (filter1 pred ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (if (pred (car ls)) (cons (car ls) res) res)))))

(define (any1 pred ls)
  (let lp ((ls ls)) (and (pair? ls) (or (pred (car ls)) (lp (cdr ls))))))

(define (expand-module-names spec)
  (cond
    ((null? spec) (list spec))
    ((not (pair? spec)) (list (list spec)))
    (else
     (append-map1
      (lambda (head)
        (map (lambda (tail) (append head tail))
             (expand-module-names (cdr spec))))
      (if (pair? (car spec))
        (append-map1 expand-module-names (car spec))
        (list (list (car spec))))))))

(define (expand-module-name-list ls)
  (append-map1 expand-module-names ls))

(define (expand-module-symbol-list ls)
  (map module-name->symbol (expand-module-name-list ls)))

(define (module-name->symbol ls)
  (if (and (eq? 'srfi (car ls)) (pair? (cdr ls)) (null? (cddr ls)))
    (string->symbol
     (string-append
      "srfi-"
      (->string (cond ((assq (cadr ls) srfi-aliases) => cdr)
                      (else (cadr ls))))))
    (string->symbol (string-intersperse (map ->string ls) "/"))))

(define (module-declarations mod)
  (let ((str (->string mod)))
    (let lp ((dirs (cons "." ##sys#include-pathnames)))
      (and (pair? dirs)
           (let ((path (string-append (car dirs) "/" str ".scm")))
             (or
              (and (file-exists? path)
                   (condition-case
                    (with-input-from-file path
                      (lambda ()
                        (let lp ()
                          (let ((x (read)))
                            (and (not (eof-object? x))
                                 (or (and (pair? x)
                                          (eq? 'common-module (car x))
                                          (caddr x))
                                     (lp)))))))
                    (e () #f)))
              (lp (cdr dirs))))))))

(define (psyntax-module? mod)
  (let ((ls (extension-info mod)))
    (if ls
      (and (assq 'module ls) #t)
      (and (module-declarations mod) #t))))

(define (module-exports mod)
  (let ((ls (extension-info mod)))
    (if ls
      (cond ((assq 'exports ls) => cdr) (else #f))
      (let lp ((ls (module-declarations mod)) (res '()))
        (cond
          ((null? ls) res)
          ((pair? (car ls))
           (case (caar ls)
             ((export export-syntax)
              (lp (cdr ls)
                  (append res (map (lambda (x) (if (pair? x) (car x) x))
                                   (cdar ls)))))
             ((inherit)
              (lp (cdr ls)
                  (append res (module-exports
                               (module-name->symbol (cadar ls))))))
             (else (lp (cdr ls) res))))
          (else (lp (cdr ls) res)))))))

(define (generate-common-module name clauses body)
  (define (expand-declarations ls)
    (let lp ((ls ls) (decls '()) (ext '()) (imps '()))
      (if (null? ls)
        (append
         (if (pair? decls) `((declare ,@decls)) '())
         (if (pair? ext) `((require-extension ,@ext)) '())
         (if (pair? imps) (map (lambda (x) `(import ,x)) imps) '()))
        (case (if (pair? (car ls)) (caar ls) 'dummy)
          ((fixnum) (lp (cdr ls) (cons '(fixnum) decls) ext imps))
          ((number-tower) (lp (cdr ls) decls (cons 'numbers ext) imps))
          ((unicode) (lp (cdr ls) decls (cons 'utf8 ext) (cons 'utf8 imps)))
          (else (lp (cdr ls) decls ext imps))))))
  (define (expand-cond clauses)
    (define (every1 pred ls)
      (not (any1 (lambda (x) (not (pred x))) ls)))
    (define (true-clause? x)
      (case x
        ((chicken else) #t)
        (else
         (and (pair? x)
              (case (car x)
                ((and) (every1 true-clause? (cdr x)))
                ((or)  (any1 true-clause? (cdr x)))
                ((not) (not (true-clause? (cadr x))))
                (else #f))))))
    (cond
      ((null? clauses) '())
      ((true-clause? (caar clauses)) (cdar clauses))
      (else (expand-cond (cdr clauses)))))
  ;;(printf "name: ~S clauses: ~S body: ~S\n" name clauses body)
  (let lp ((ls clauses) (decls '()) (imps '()) (macs '()) (ext '())
           (exps '()) (main #f) (mods '()) (body body))
    (if (pair? ls)
      (case (if (pair? (car ls)) (caar ls) 'dummy)
        ((declare)
         (lp (cdr ls) (append (cdar ls) decls) imps macs ext exps mods body))
        ((export export-syntax)
         (lp (cdr ls) decls imps macs ext
             (append (map (lambda (x) (if (pair? x) (car x) x)) (cdar ls)) exps)
             main mods body))
        ((import)
         (lp (cdr ls) decls
             (append imps (expand-module-symbol-list (cdar ls)))
             macs ext exps main mods body))
        ((import-extension)
         (let* ((new-exts (expand-module-symbol-list (cdar ls)))
                (new-imps (filter1 psyntax-module? new-exts)))
           (lp (cdr ls) decls (append imps new-imps) macs (append ext new-exts)
               exps main mods body)))
        ((import-lazy)
         (let ((new-exts (expand-module-symbol-list (cadar ls)))
               (new-imps (filter1 psyntax-module? new-exts)))
           (lp (cdr ls) decls (append imps mods) macs (append ext new-exts)
               exps main mods body)))
        ((import-rename)
         (let* ((mod-name (cadar ls))
                (mod-sym (module-name->symbol mod-name))
                (local-name (caddar ls))
                (exported-name (car (cdddar ls))))
           (if (not (psyntax-module? mod-sym))
             (lp (cdr ls) decls imps macs (append ext (list mod-sym))
                 exps main
                 (append mods `((define ,local-name ,exported-name))) body)
             (let* ((tmp-name (gensym 'temp-mod))
                    (tmp-mod
                     `(module ,tmp-name (,local-name)
                              (import ,mod-sym)
                              (define ,local-name ,exported-name))))
               (lp (cdr ls) decls (append imps (list tmp-name))
                   macs (append ext (list mod-sym)) exps main
                   (append mods (list tmp-mod)) body)))))
        ((import-prefix)
         (let* ((mod-name (cadar ls))
                (mod-sym (module-name->symbol mod-name))
                (prefix-string (if (keyword? (caddar ls))
                                 (string-append
                                  (keyword->string (caddar ls))
                                  ":")
                                 (->string (caddar ls))))
                (prefix
                 (lambda (x)
                   (string->symbol
                    (string-append prefix-string (->string x)))))
                (exports (module-exports mod-sym))
                (prefixed-exports
                 (map prefix exports))
                (prefixed-decls
                 (map (lambda (a b) `(define ,a ,b)) prefixed-exports exports)))
           (if (not (psyntax-module? mod-sym))
             (lp (cdr ls) decls imps macs (append ext (list mod-sym))
                 exps main (append mods prefixed-decls) body)
             (let* ((tmp-name (gensym 'temp-mod))
                    (tmp-mod
                     `(module ,tmp-name ,prefixed-exports
                              (import ,mod-sym)
                              ,@prefixed-decls)))
               (lp (cdr ls) decls (append imps (list tmp-name))
                   macs (append ext (list mod-sym)) exps main
                   (append mods (list tmp-mod)) body)))))
        ((import-syntax)
         (lp (cdr ls) decls imps (append (cdar ls) macs)
             ext exps main mods body))
        ((inherit)
         (let* ((mod-syms (expand-module-symbol-list (cdar ls)))
                (new-imps (filter1 psyntax-module? mod-syms))
                (exports (append-map1 module-exports mod-syms)))
           (lp (cdr ls) decls (append imps new-imps) macs
               (append ext mod-syms) (append exps exports) main mods body)))
        ((entry-point)
         (lp (cdr ls) decls imps macs ext exps (cadar ls) mods body))
        ((cond)
         (lp (append (expand-cond (cdar ls)) (cdr ls))
             decls imps macs ext exps main mods body))
        ((prelude postlude prelude-include postlude-include)
         (lp (cdr ls) decls imps macs ext exps main mods
             ((case (caar ls)
                ((prelude prelude-include) append)
                (else (lambda (a b) (append b a))))
              (case (caar ls)
                ((prelude postlude) (cdar ls))
                (else `((include ,(cadar ls)))))
              body)))
        (else
         (lp (cdr ls) decls imps macs ext exps main mods body)))
      (let ((macs (expand-module-symbol-list macs)))
        (for-each
         (lambda (m) (require m) (if (psyntax-module? m) (eval `(import ,m))))
         macs)
        ;;(fprintf (current-error-port) "imps: ~S ext: ~S mods: ~S\n" imps ext mods)
        `(begin
           ,@(expand-declarations decls)
           ,@(if (pair? ext) `((require-extension ,@ext)) '())
           ,@(map (lambda (m) `(require-extension ,m)) macs)
           (module
            ,@(if (null? name) name (list (module-name->symbol name)))
            ,exps
            ,@mods
            ,@(map (lambda (m) `(import ,m)) imps)
            ,@body
            ,@(if main
                `((cond-expand
                   (compiling
                    (,main (argv)))
                   (else
                    (,main (cons (car (argv)) (command-line-arguments))))))
                '())))))))
