
(require-extension (srfi 10000))

(common-module (common combinators)
  ((export identity true false constantly always-true always-false
           complement compose flip conjoin disjoin fix noop S K I U Y))

(define true #t)
(define false #f)

(define (constantly x) (lambda _ x))

(define (identity x) x)

(define always-true (constantly #t))
(define always-false (constantly #f))

(define (complement f) (lambda (x) (not (f x))))

(define (flip f)
  (lambda (x y) (f y x)))

(define (make-chain-compiler combine2 empty)
  (define (combine . o)
    (if (null? o)
      empty
      (let ((o2 (cdr o)))
        (if (null? o2)
          (car o)
          (if (null? (cdr o2))
            (combine2 (car o) (car o2))
            (combine2 (car o) (apply combine o2)))))))
  combine)

(define (conjoin2 f g)
  (cond
    ((eq? f identity) g)
    ((eq? g identity) f)
    ((eq? f always-true) g)
    ((eq? f always-false) always-false)
    ((eq? g always-true) f)
    ((eq? g always-false) always-false)
    (else (lambda (x) (and (f x) (g x))))))

(define conjoin (make-chain-compiler conjoin2 always-true))

(define (disjoin2 f g)
  (cond
    ((eq? f identity) g)
    ((eq? g identity) f)
    ((eq? f always-true) always-true)
    ((eq? f always-false) g)
    ((eq? g always-true) always-true)
    ((eq? g always-false) f)
    (else (lambda (x) (or (f x) (g x))))))

(define disjoin (make-chain-compiler disjoin2 always-false))

(define (compose2 f g)
  (cond
    ((eq? f identity) g)
    ((eq? g identity) f)
    ((eq? f always-true) always-true)
    ((eq? f always-false) always-false)
    ((eq? g always-true) (constantly (f true)))
    ((eq? g always-false) (constantly (f false)))
    (else (lambda (x) (f (g x))))))

(define compose (make-chain-compiler compose2 identity))

(define (noop . _) (if #f #f))

(define (fix proc init . opt)
  (let ((eq (if (pair? opt) (car opt) equal?)))
    (let loop ((x (proc init))
               (last init))
      (if (eq x last)
        x
        (loop (proc x) x)))))

(define (S f g x) ((f x) (g x)))
(define (K x y) x)
(define (I x) x)
(define (U x) (x x))
(define (Y gen)
  ((lambda (f) (f f))
   (lambda (f)
     (gen (lambda x (apply (f f) x))))))

)
