
(require-extension (srfi 10000))

(common-module (common file)
  ((export file-regular? file-directory? file-readable? file-writable?
           directory-fold directory-fold-tree copy-file create-directory*
           find-in-path delete-file))

;;file-exists?
;;file-size

(define (file-regular? file)
  (eq? 'regular (file-info-type (file-info file))))

(define (file-directory? file)
  (eq? 'directory (file-info-type (file-info file))))

(define (make-file-permission-predicate user group world)
  (lambda (file)
    (let* ((info (file-info file))
           (mode (file-info-mode info)))
      (or (not (zero? (bitwise-and mode world)))
          (and (not (zero? (bitwise-and mode user)))
               (eqv? (file-info-owner info)
                     (user-info-uid (user-info (getenv "USER")))))
          (and (not (zero? (bitwise-and mode group)))
               (eqv? (file-info-group info)
                     (user-info-gid (user-info (getenv "USER")))))))))

(define file-readable? (make-file-permission-predicate #o400 #o040 #o004))
(define file-writable? (make-file-permission-predicate #o200 #o020 #o002))

(define (file-modification-time f)
  (time->seconds (file-last-modification-time f)))

(define (copy-file from to)
  (shell-command (string-append "cp " from " " to)))

(define delete-file
  (let ((orig-delete-file delete-file))
    (lambda (f)
      (with-exception-handler
          (lambda (e) (delete-directory f))
        (lambda () (orig-delete-file f))))))

(define (create-directory* f)
  (let* ((end (- (string-length f) 1))
         (f/ (if (eqv? #\/ (string-ref f end)) (substring f 0 end) f)))
    (or (file-exists? f/)
        (with-exception-handler
            (lambda (e)
              (let ((dir (path-directory f/)))
                (and (string? dir)
                     (not (string=? dir f/))
                     (create-directory* dir)
                     (create-directory f/))))
          (lambda () (create-directory f/))))))

(define (directory-fold dir kons knil)
  (let ((p (open-directory dir)))
    (let lp ((acc knil))
      (let ((s (read p)))
        (if (string? s)
          (lp (kons s acc))
          (begin
            (close-input-port p)
            acc))))))

(define (directory-fold-tree dir kons knil)
  (let lp ((ls (directory-files dir))
           (d dir)
           (hl (- (file-number-of-links dir) 2))
           (acc knil))
    (cond
      ((null? ls)
       acc)
      ((and (positive? hl)
            (let ((d2 (string-append d "/" (car ls))))
              (and (file-directory? d2) d2)))
       =>
       (lambda (d2)
         (lp (cdr ls)
             d
             (- hl 1)
             (kons d (car ls)
                   (lp (directory-files d2)
                       d2
                       (- (file-number-of-links dir) 2)
                       acc)))))
      (else
       (lp (cdr ls) d hl (kons d (car ls) acc))))))

(define (find-in-path file ls . o)
  (let ((pred (if (pair? o) (car o) (lambda (f) #t))))
    (let lp ((ls ls))
      (and (pair? ls)
           (let* ((dir (car ls))
                  (len (string-length dir))
                  (path (if (eqv? #\/ (string-ref dir (- len 1)))
                          (string-append dir file)
                          (string-append dir "/" file))))
             (if (and (file-exists? path) (pred path))
               path
               (lp (cdr ls))))))))

)

