
;(load "~~/syntax-case")

; (define-syntax (define-macro stx)
;   (syntax-case stx ()
;    ((_ (macro . args) . body)
;     (syntax (define-macro macro (lambda args . body))))
;    ((_ macro transformer)
;     (syntax
;      (define-syntax (macro stx2)
;        (let ((v (syntax-object->datum stx2)))
;          (datum->syntax-object
;           ; we need the *identifier* of the macro call  
;           ; (there is probably a smarter way of extracting that ...)
;           (syntax-case stx2 () ((name . more) (syntax name)))
;           (apply transformer (cdr v)))))))))

(define-macro (require-extension . o) #f)

(define-macro (let-optionals* ls params . body)
  `(apply (lambda (#!optional ,@params) ,@body) ,ls))

(define-macro (let-keywords* ls params . body)
  `(apply (lambda (#!key ,@params) ,@body) ,ls))

(define-macro (common-module name clauses . body)

(define *load-path* '("." "~~/modules"))

(define (find-in-path file ls)
  (let lp ((ls ls))
    (and (pair? ls)
         (let ((f (string-append (car ls) "/" file)))
           (if (file-exists? f)
             f
             (lp (cdr ls)))))))

; (define (module->path sym)
;   (let* ((s (symbol->string sym))
;          (path
;           (or (find-in-path (string-append s ".o1") *load-path*)
;               (find-in-path (string-append s ".scm") *load-path*))))
;     (and path (path-strip-extension path))))

(define srfi-aliases
  '((cond-expand . 0)
    (list-lib . 1)
    (and-let* . 2)
    (numeric-vectors . 4)
    (signature-let . 5)
    (string-ports . 6)
    (configuration-language . 7)
    (receive . 8)
    (define-record-type . 9)
    (external-form-syntax . 10)
    (multiple-value-bind . 11)
    (string-lib . 13)
    (char-set-lib . 14)
    (case-lambda . 16)
    (generalized-set! . 17)
    (multithreading . 18)
    (time-lib . 19)
    (real-time-multithreading . 21)
    (unix-scheme-scripts . 22)
    (error . 23)
    (array-lib . 25)
    (cut . 26)
    (random-numbers . 27)
    (basic-format . 28)
    (localization . 29)
    (multi-line-comment-syntax . 30)
    (rec . 31)
    (sort-lib . 32)
    (bitwise-lib . 33)
    (exceptions . 34)
    (conditions . 35)
    (i/o-conditions . 36)
    (args-fold . 37)
    (shared-structure-read/write . 38)
    (parameters . 39)
    (streams-lib . 40)
    (comprehensions . 42)
    (vector-lib . 43)
    (comprehensions . 44)
    (lazy-eval . 45)
    (syntax-rules-choose-ellipse . 46)
    (homogeneous-array-lib . 47)
    (intermediate-format . 48)
    (require-extension . 55)
    (advanced-records . 57)
    (array-notation . 58)
    (vicinity . 59)
    (bitwise-operations . 60)
    (sexpr-comments . 62)
    (homogeneous-and-heterogeneous-arrays . 63)
    (testing . 64)
    ))

(define (append-map1 f ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (append (reverse (f (car ls))) res)))))

(define (filter1 pred ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (if (pred (car ls)) (cons (car ls) res) res)))))

(define (any1 pred ls)
  (let lp ((ls ls)) (and (pair? ls) (or (pred (car ls)) (lp (cdr ls))))))

(define (expand-module-names spec)
  (cond
    ((null? spec) (list spec))
    ((not (pair? spec)) (list (list spec)))
    (else
     (append-map1
      (lambda (head)
        (map (lambda (tail) (append head tail))
             (expand-module-names (cdr spec))))
      (if (pair? (car spec))
        (append-map1 expand-module-names (car spec))
        (list (list (car spec))))))))

(define (x->string x)
  (cond ((string? x) x)
        ((symbol? x) (symbol->string x))
        ((number? x) (number->string x))
        (else (object->string x))))

(define (module-name->symbol ls)
  (define (string-intersperse ls str)
    (if (null? ls)
      ""
      (let lp ((ls (cdr ls)) (res (list (car ls))))
        (if (null? ls)
          (apply string-append (reverse res))
          (lp (cdr ls) (cons (car ls) (cons str res)))))))
  (if (and (eq? 'srfi (car ls)) (pair? (cdr ls)) (null? (cddr ls)))
    (string->symbol
     (string-append
      "srfi-"
      (x->string (cond ((assq (cadr ls) srfi-aliases) => cdr)
                       (else (cadr ls))))))
    (string->symbol (string-intersperse (map x->string ls) "/"))))

(define (expand-cond clauses)
  (define (every1 pred ls)
    (not (any1 (lambda (x) (not (pred x))) ls)))
  (define (true-clause? x)
    (case x
      ((gambit else) #t)
      (else
       (and (pair? x)
            (case (car x)
              ((and) (every1 true-clause? (cdr x)))
              ((or)  (any1 true-clause? (cdr x)))
              ((not) (not (true-clause? (cadr x))))
              (else #f))))))
  (cond
    ((null? clauses) '())
    ((true-clause? (caar clauses)) (cdar clauses))
    (else (expand-cond (cdr clauses)))))

(define (module-declarations mod)
  (let ((str (x->string mod)))
    (let lp ((dirs *load-path*))
      (and (pair? dirs)
           (let ((path (string-append (car dirs) "/" str ".scm")))
             (or
              (and (file-exists? path)
                   (with-exception-handler
                       (lambda (e) #f)
                     (lambda ()
                       (with-input-from-file path
                         (lambda ()
                           (let lp ()
                             (let ((x (read)))
                               (and (not (eof-object? x))
                                    (or (and (pair? x)
                                             (eq? 'common-module (car x))
                                             (caddr x))
                                        (lp))))))))))
              (lp (cdr dirs))))))))

(define (module-exports mod)
  (let lp ((ls (module-declarations mod)) (res '()))
    (cond
      ((null? ls) res)
      ((pair? (car ls))
       (case (caar ls)
         ((export export-syntax)
          (lp (cdr ls)
              (append res (map (lambda (x) (if (pair? x) (car x) x))
                               (cdar ls)))))
         ((inherit)
          (lp (cdr ls)
              (append res (module-exports
                           (module-name->symbol (cadar ls))))))
         (else (lp (cdr ls) res))))
      (else (lp (cdr ls) res)))))

(define (generate-common-module name clauses body)
  (let lp ((ls clauses) (decls '()) (imps '()) (macs '()) (exps '()) (main #f) (body body))
    (if (pair? ls)
      (case (if (pair? (car ls)) (caar ls) 'dummy)
        ((declare)
         (lp (cdr ls) (append (cdar ls) decls) imps macs exps main body))
        ((export export-syntax)
         (lp (cdr ls) decls imps macs
             (append (map (lambda (x) (if (pair? x) (car x) x)) (cdar ls)) exps)
             main body))
        ((import import-extension import-lazy)
         (lp (cdr ls) decls (append (cdar ls) imps) macs exps main body))
        ((import-syntax)
         (lp (cdr ls) decls imps (append (cdar ls) macs) exps main body))
        ((import-rename)
         (let* ((mod-name (cadar ls))
                (local-name (caddar ls))
                (exported-name (car (cdddar ls))))
           (lp (cdr ls) decls (cons mod-name imps) macs exps main
               (cons `(define ,local-name ,exported-name) body))))
        ((import-prefix)
         (let* ((mod-name (cadar ls))
                (mod-sym (module-name->symbol mod-name))
                (prefix-string (if (keyword? (caddar ls))
                                 (string-append
                                  (keyword->string (caddar ls))
                                  ":")
                                 (x->string (caddar ls))))
                (prefix
                 (lambda (x)
                   (string->symbol
                    (string-append prefix-string (x->string x))))))
           (lp (cdr ls) decls (cons mod-name imps) macs exps main
               (append
                (map (lambda (x) `(define ,(prefix x) ,x))
                     (module-exports mod-sym))
                body))))
        ((inherit)
         (lp (cdr ls) decls (append (cdar ls) imps) macs exps main body))
        ((entry-point)
         (lp (cdr ls) decls imps macs exps (cadar ls) body))
        ((cond)
         (lp (append (expand-cond (cdar ls)) (cdr ls))
             decls imps macs exps main body))
        ((prelude postlude prelude-include postlude-include)
         (lp (cdr ls) decls imps macs exps main
             ((case (caar ls)
                ((prelude prelude-include) append)
                (else (lambda (a b) (append b a))))
              (case (caar ls)
                ((prelude postlude) (cdar ls))
                (else `((include ,(find-in-path (cadar ls) *load-path*)))))
              body)))
        (else (lp (cdr ls) decls imps macs exps main body)))
      (let ((imps (reverse (map module-name->symbol
                                   (append-map1 expand-module-names imps))))
            (macs (map module-name->symbol
                         (append-map1 expand-module-names macs))))
        `(begin
           ,@(if main '((load "~~/require")) '())
           ,@(map (lambda (m)
                    `(begin
                       (include ,(find-in-path
                                  (string-append
                                   (symbol->string m) ".scm") *load-path*))
                       (require ',m)))
                  macs)
           ,@(map (lambda (m) `(require ',m)) imps)
           ,@body
           ,@(if main `((,main (command-line))) '()))))))

  (let ((res (generate-common-module name clauses body)))
    ;;(display "body: ") (write res) (newline)
    res))

; (define-macro (import-extension . args)
;   `(begin
;      ,@(map
;         (lambda (m) `(require ',m))
;         (map module-name->symbol
;              (append-map1 expand-module-names args)))))

; (define-macro (import-syntax . args)
;   `(begin
;      ,@(map
;         (lambda (m) `(include ,(string-append (symbol->string m) ".scm")))
;         (map module-name->symbol
;              (append-map1 expand-module-names args)))))

; (define-macro (import-lazy module . procs)
;   `(import-extension ,module))
