
(define-module common.file
  (extend file.util)
  (export-all))
(select-module common.file)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;file-exists?

(define file-directory? file-is-directory?)
(define file-regular? file-is-regular?)
(define file-readable? file-is-readable?)
(define file-writable? file-is-writable?)

(define file-modification-time file-mtime)

;;file-size

(define (directory-files f)
  (directory-list f :children? #t))

(define (directory-fold dir kons knil)
  (let lp ((ls (cddr (sys-readdir dir))) (acc knil))
    (if (null? ls)
      acc
      (lp (cdr ls) (kons (car ls) acc)))))

(define (directory-fold-tree dir kons knil)
  (let lp ((ls (directory-files dir))
           (d dir)
           (hl (- (file-nlink dir) 2))
           (acc knil))
    (cond
      ((null? ls)
       acc)
      ((and-let* (((positive? hl))
                  (d2 (string-append d "/" (car ls)))
                  ((file-directory? d2)))
         d2)
       =>
       (lambda (d2)
         (lp (cdr ls)
             d
             (- hl 1)
             (kons d (car ls)
                   (lp (directory-files d2)
                       d2
                       (- (file-nlink d2) 1)
                       acc)))))
      (else
       (lp (cdr ls) d hl (kons d (car ls) acc))))))

(define (directory-files dir)
  (cddr (sys-readdir dir)))

(define (create-directory d)
  (sys-mkdir d #o755))

; (define (delete-file f)
;   (with-exception-handler
;       (lambda _ (sys-rmdir f))
;     (lambda () (sys-unlink f))))

(define (delete-file f)
  (if (file-directory? f)
    (sys-rmdir f)
    (sys-unlink f)))

(define rename-file sys-rename)

(define (find-in-path file ls . o)
  (let ((pred (if (pair? o) (car o) (lambda (f) #t))))
    (let lp ((ls ls))
      (and (pair? ls)
           (let* ((dir (car ls))
                  (len (string-length dir))
                  (path (if (eqv? #\/ (string-ref dir (- len 1)))
                          (string-append dir file)
                          (string-append dir "/" file))))
             (if (and (file-exists? path) (pred path))
               path
               (lp (cdr ls))))))))

(provide "common/file")
