
(define-module common.pathname
  (use file.util)
  (use srfi-13)
  (export path-absolute? path-strip-directory path-directory
          path-extension path-strip-extension path-replace-extension
          path-expand path-normalize path-relative path-build
          current-directory set-current-directory!))
(select-module common.pathname)

(define current-directory current-directory)
(define set-current-directory! sys-chdir)

(define path-absolute? absolute-path?)

(define (path-strip-directory x)
  (if (and (not (string=? x ""))
           (eqv? #\/ (string-ref x (- (string-length x) 1))))
    ""
    (sys-basename x)))

(define path-directory sys-dirname)

;; Return #f or ""?  I like #f so you can more easily use this in
;; conditionals and can always just write (or (path-extension f) "").

(define path-extension
  (let ((orig-path-extension path-extension))
    (lambda (path)
      (let ((res (orig-path-extension path)))
        (and (not (equal? res "")) res)))))

;; Headaches...
;;
;; We chose POSIX compatibility, Gauche chose compatibility with Perl's
;; File::Basename here on the border case (path-strip-extension "/").
;; The File::Basename manpage itself mentions compatibility with
;; basename(1), however it doesn't match the shell command behavior on
;; this case.  What File::Basename is actually being compatible with is
;; the GNU version of basename(3).  Gambit also chooses compatibility
;; with GNU over POSIX, so I may need to rethink this.

(define (path-strip-extension path)
  (let ((i (string-index-right path (lambda (c) (or (eqv? c #\.) (eqv? c #\/))))))
    (if (and i
             (< (+ i 1) (string-length path))
             (not (eqv? (string-ref path i) #\/)))
      (substring path 0 i)
      path)))

(define (path-replace-extension f ext)
  (string-append (path-strip-extension f) "." ext))

(define path-normalize simplify-path)

(define (path-expand f . o)
  (let ((res (expand-path f)))
    (if (absolute-path? res)
      res
      (path-build (if (pair? o) (car o) (current-directory)) res))))

(define (path-build . o)
  (let lp ((ls o) (res '()))
    (cond ((null? ls) (apply build-path (reverse res)))
          ((pair? (car ls)) (lp (append (car ls) (cdr ls)) res))
          (else (lp (cdr ls) (cons (x->string (car ls)) res))))))

(define (path-relative path . o)
  (let* ((path (path-normalize (path-expand path)))
         (p-len (string-length path))
         (dir (path-normalize (if (or (null? o) (equal? (car o) "."))
                                (current-directory)
                                (path-expand (car o)))))
         (d-len (string-length dir)))
    (and (> p-len d-len)
         (string=? dir (substring path 0 d-len))
         (cond
           ((eqv? #\/ (string-ref path d-len))
            (substring path (+ d-len 1) p-len))
           ((eqv? #\/ (string-ref path (- d-len 1)))
            (substring path d-len p-len))
           (else #f)))))

(provide "common/pathname")
