
(define-module common.process
  (use gauche.process)
  (export process-run process-run-and-exit
          process-output->string process-output->string-list
          with-input-from-process with-output-to-process with-process-io
          shell-escape))
(select-module common.process)

(define (shell-escape-char? c)
  (not (or (char-alphabetic? c)
           (char-numeric? c)
           (memv c '(#\_ #\- #\+ #\: #\, #\= #\/ #\.)))))

(define (shell-escape-backslash str)
  (with-output-to-string
    (lambda ()
      (with-input-from-string str
        (lambda ()
          (let lp ()
            (let ((c (read-char)))
              (if (not (eof-object? c))
                (begin
                  (if (shell-escape-char? c)
                    (display #\\))
                  (display c)
                  (lp))))))))))

(define (shell-escape str)
  (let lp ((i (- (string-length str) 1)) (quote? #f))
    (if (negative? i)
      (if quote?
        (string-append "\"" str "\"")
        str)
      (let ((c (string-ref str i)))
        (cond
          ((or (eqv? c #\") (eqv? c #\\))
           (shell-escape-backslash str))
          ((shell-escape-char? c)
           (lp (- i 1) #t))
          (else
           (lp (- i 1) quote?)))))))

(define (make-process-string command ls)
  (define (escape x)
    (shell-escape (if (string? x) x (x->string x))))
  (string-join (cons command (map escape ls))))

(define (process-run x . args)
  (zero? (sys-system (make-process-string x args))))

(define (process-run-and-exit x . args)
  (sys-exec (make-process-string x args) '()))

(define orig-with-input-from-process with-input-from-process)
(define orig-with-output-to-process with-output-to-process)

(define (with-input-from-process x thunk)
  (orig-with-input-from-process
   (if (pair? x) (make-process-string (car x) (cdr x)) x)
   thunk))

(define (with-output-to-process x thunk)
  (orig-with-output-to-process
   (if (pair? x) (make-process-string (car x) (cdr x)) x)
   thunk))

(define (with-process-io x thunk)
  (call-with-process-io
      (if (pair? x) (make-process-string (car x) (cdr x)) x)
    (lambda (in out)
      (with-input-from-port in
        (lambda ()
          (with-output-to-port out
            thunk))))))

(define (process-output->string-list x . args)
  (let ((command (make-process-string x args)))
    (with-input-from-process command
      (lambda ()
        (let lp ((res '()))
          (let ((line (read-line)))
            (if (eof-object? line)
              (reverse res)
              (lp (cons line res)))))))))

(define (process-output->string . o)
  (string-join (apply process-output->string-list o) "\n"))

(provide "common/process")
