
(define-module srfi-10000
  (export common-module import-extension import-syntax import-lazy let-keywords*))
(select-module srfi-10000)

(define-macro (common-module name clauses . body)
  (generate-common-module
   (if (null? name) (gensym "%anonymous-module") (module-name->symbol name))
   clauses
   body))

(define-macro (import-extension . args)
  `(use ,@(expand-module-symbol-list args)))

(define-macro (import-syntax . args)
  `(import-extension ,@args))

(define-macro (import-lazy mod . args)
  `(autoload ,@(module-name->symbol (car (expand-module-names mod))) ,@args))

(define srfi-aliases
  '((cond-expand . 0)
    (list-lib . 1)
    (and-let* . 2)
    (numeric-vectors . 4)
    (signature-let . 5)
    (string-ports . 6)
    (configuration-language . 7)
    (receive . 8)
    (define-record-type . 9)
    (external-form-syntax . 10)
    (multiple-value-bind . 11)
    (string-lib . 13)
    (char-set-lib . 14)
    (case-lambda . 16)
    (generalized-set! . 17)
    (multithreading . 18)
    (time-lib . 19)
    (real-time-multithreading . 21)
    (unix-scheme-scripts . 22)
    (error . 23)
    (array-lib . 25)
    (cut . 26)
    (random-numbers . 27)
    (basic-format . 28)
    (localization . 29)
    (multi-line-comment-syntax . 30)
    (rec . 31)
    (sort-lib . 32)
    (bitwise-lib . 33)
    (exceptions . 34)
    (conditions . 35)
    (i/o-conditions . 36)
    (args-fold . 37)
    (shared-structure-read/write . 38)
    (parameters . 39)
    (streams-lib . 40)
    (comprehensions . 42)
    (vector-lib . 43)
    (comprehensions . 44)
    (lazy-eval . 45)
    (syntax-rules-choose-ellipse . 46)
    (homogeneous-array-lib . 47)
    (intermediate-format . 48)
    (require-extension . 55)
    (advanced-records . 57)
    (array-notation . 58)
    (vicinity . 59)
    (bitwise-operations . 60)
    (sexpr-comments . 62)
    (homogeneous-and-heterogeneous-arrays . 63)
    (testing . 64)
    ))

(define (append-map1 f ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (append (reverse (f (car ls))) res)))))

(define (expand-module-names spec)
  (cond
    ((null? spec) (list spec))
    ((not (pair? spec)) (list (list spec)))
    (else
     (append-map1
      (lambda (head)
        (map (lambda (tail) (append head tail))
             (expand-module-names (cdr spec))))
      (if (pair? (car spec))
        (append-map1 expand-module-names (car spec))
        (list (list (car spec))))))))

(define (expand-module-name-list ls)
  (append-map1 expand-module-names ls))

(define (expand-module-symbol-list ls)
  (map module-name->symbol (expand-module-name-list ls)))

(define (module-name->symbol ls)
  (if (and (eq? 'srfi (car ls)) (pair? (cdr ls)) (null? (cddr ls)))
    (string->symbol
     (string-append
      "srfi-"
      (x->string (cond ((assq (cadr ls) srfi-aliases) => cdr)
                       (else (cadr ls))))))
    (string->symbol (string-join (map x->string ls) "."))))

(define (generate-common-module name clauses body)
  (define (expand-cond clauses)
    (define (every1 pred ls)
      (not (any (lambda (x) (not (pred x))) ls)))
    (define (true-clause? x)
      (case x
        ((gauche else) #t)
        (else
         (and (pair? x)
              (case (car x)
                ((and) (every1 true-clause? (cdr x)))
                ((or)  (any true-clause? (cdr x)))
                ((not) (not (true-clause? (cadr x))))
                (else #f))))))
    (cond
      ((null? clauses) '())
      ((true-clause? (caar clauses)) (cdar clauses))
      (else (expand-cond (cdr clauses)))))
  (define (expand-imports ls)
    (let lp ((ls ls) (res '()))
      (cond
        ((null? ls)
         (reverse res))
        ((eq? (caar ls) 'import-lazy)
         (lp (cdr ls)
             (cons (cons 'autoload
                         (cons (module-name->symbol
                                (car (expand-module-names (list (cadar ls)))))
                               (cddar ls)))
                   res)))
        ((memq (caar ls) '(import import-only import-extension import-syntax))
         (let ((im (if (memq (caar ls) '(import import-only)) 'import 'use)))
           (lp (cdr ls)
               (append (map
                        (lambda (x) (list im (module-name->symbol x)))
                        (expand-module-name-list (cdar ls)))
                       res))))
        (else
         (lp (cdr ls) res)))))
  (let lp ((ls clauses)
           (imps '())
           (exps '())
           (main #f)
           (mods '())
           (body body))
    (if (pair? ls)
      (case (if (pair? (car ls)) (caar ls) 'dummy)
        ((entry-point)
         (lp (cdr ls) imps exps (cadar ls) mods body))
        ((export export-syntax)
         (lp (cdr ls) imps (append (cdar ls) exps) main mods body))
        ((keywords)
         (let* ((syms (cdar ls))
                (keys
                 (map
                  (lambda (s)
                    (let ((str (symbol->string s)))
                      (make-keyword
                       (substring str 0 (- (string-length str) 1)))))
                  syms)))
           ;;(lp (cdr ls) imps (append exps syms) main mods
           ;;    (append (map (lambda (s k) `(define ,s ,k)) syms keys) body))
           (lp (cdr ls) imps (append exps syms) main mods
               (append (map (lambda (s) `(define ,s ',s)) (cdar ls)) body))
           ))
        ((import import-only import-extension import-syntax import-lazy)
         (lp (cdr ls) (cons (car ls) imps) exps main mods body))
        ((import-rename)
         (let* ((mod-name (cadar ls))
                (mod-sym (module-name->symbol mod-name))
                (local-name (caddar ls))
                (exported-name (car (cdddar ls))))
           (load (module-name->path mod-sym))
           (lp (cdr ls) imps exps main mods
               (cons `(define ,local-name
                        (with-module ,mod-sym ,exported-name))
                     body))))
        ((import-prefix)
         (let* ((mod-name (cadar ls))
                (mod-sym (module-name->symbol mod-name))
                (prefix-string (if (keyword? (caddar ls))
                                 (string-append
                                  (keyword->string (caddar ls))
                                  ":")
                                 (x->string (caddar ls))))
                (prefix
                 (lambda (x)
                   (string->symbol
                    (string-append prefix-string (x->string x))))))
           (load (module-name->path mod-sym))
           (let ((exports (module-exports (find-module mod-sym))))
             (lp (cdr ls) imps exps main mods
                 (append
                  (map
                   (lambda (x)
                     `(define ,(prefix x) (with-module ,mod-sym ,x)))
                   exports)
                  body)))))
        ((inherit)
         (lp (cdr ls) imps exps main mods
             `((extend ,@(expand-module-symbol-list (cdar ls))) ,@body)))
        ((cond)
         (lp (append (expand-cond (cdar ls)) (cdr ls))
             imps exps main mods body))
        ((prelude postlude prelude-include postlude-include)
         (lp (cdr ls) imps exps main mods
             ((case (caar ls)
                ((prelude prelude-include) append)
                (else (lambda (a b) (append b a))))
              (case (caar ls)
                ((prelude postlude) (cdar ls))
                (else `((load ,(cadar ls)))))
              body)))
        (else
         (lp (cdr ls) imps exps main mods body)))
      (let ((mod
             `(define-module ,name
                (use srfi-10000)
                ,@(expand-imports imps)
                (export ,@exps)
                ,@body
                (provide ,(module-name->path name)))))
        (if (or main (pair? mods))
          `(begin
             ,@mods
             ,mod
             ,@(if main
                 `(((with-module ,name ,main)
                    (with-module user (cons *program-name* *argv*))))
                 '()))
          mod)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (proper-list? x)
  (or (null? x) (and (pair? x) (proper-list? (cdr x)))))

(define (with-split-improper-list ls proc)
  (let loop ((l ls) (res '()))
    (cond ((pair? l) (loop (cdr l) (cons (car l) res)))
          (else (proc (reverse! res) l)))))

(define (with-split-at-sym sym ls proc)
  (let lp ((ls ls) (head '()))
    (if (or (null? ls) (eq? sym (car ls)))
      (proc (reverse head) ls)
      (lp (cdr ls) (cons (car ls) head)))))

(define (x->keyword x)
  (let ((str (x->string x)))
    (string->symbol (string-append str ":"))))

(define (extract-spec x l)
  (cond
    ((symbol? x)
     (extract-spec (list x) l))
    ((and (list? x) (<= 1 (length x) 4) (symbol? (car x)))
     (let ((var (car x)))
       (with-split-at-sym '=> (cdr x)
         (lambda (info act)
           (let ((keys (if (and (pair? info) (pair? (cdr info)))
                         (list (car info))
                         (list (x->keyword var))))
                 (default (if (pair? info)
                            (if (pair? (cdr info)) (cadr info) (car info))
                            #f)))
             (list var default keys
                   (cond ((pair? act)
                          (if (not (null? (cddr act)))
                            (errorf "multiple exprs found after =>: ~S" x))
                          `(,(cadr act) (cadr ,l)))
                         (else
                          `(set! ,var (cadr ,l))))))))))
    (else (error "bad binding in let-plist*:" x))))

(define-macro (let-keywords* ls vars . body)
  (let ((undef (gensym))
        (lp (gensym))
        (l (gensym)))
    (if (proper-list? vars)
      (let ((var&default&keys&act (map (lambda (v) (extract-spec v l)) vars)))
        ;; simple loop
        `(let ((,undef (list 'undef)))
           (let (,@(map (lambda (x) `(,(car x) ,undef)) var&default&keys&act))
             (let ,lp ((,l ,ls))
               (cond
                 ((not (null? ,l))
                  (case (car ,l)
                    ,@(map cddr var&default&keys&act))
                  (,lp (cddr ,l)))
                 (else
                  ,@(map (lambda (x)
                           `(if (eq? ,(car x) ,undef)
                              (set! ,(car x) ,(cadr x))))
                         var&default&keys&act)
                  ,@body))))))
      (with-split-improper-list vars
        (lambda (vars2 others-var)
          (let ((var&default&keys&act (map (lambda (v) (extract-spec v l)) vars2)))
            ;; lp accumulating unknown keywords
            `(let ((,undef (list 'undef)))
               (let (,@(map (lambda (x) `(,(car x) ,undef)) var&default&keys&act))
                 (let ,lp ((,l ,ls) (acc '()))
                   (cond
                     ((null? ,l)
                      ,@(map (lambda (x)
                               `(if (eq? ,(car x) ,undef)
                                  (set! ,(car x) ,(cadr x))))
                             var&default&keys&act)
                      (let ((,others-var (reverse! acc)))
                        ,@body))
                     (else
                      (,lp (cddr ,l)
                          (case (car ,l)
                            ,@(map (lambda (x) (append (cddr x) '(acc)))
                                   var&default&keys&act)
                            (else
                             (cons (cadr ,l) (cons (car ,l) acc))))))))))))))))

(provide "srfi-10000")

