
(define-module (common file))

(export delete-file file-exists? file-regular? file-directory?
        file-readable? file-writable? file-modification-time
        rename-file create-directory create-directory* delete-directory
        ;;create-symbolic-link read-symbolic-link
        directory-files directory-fold directory-fold-tree
        find-in-path file-size)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;delete-file

;;file-exists?

;;rename-file

(define create-directory mkdir)

(define delete-directory rmdir)

(define create-symbolic-link symlink)

(define read-symbolic-link readlink)

(define file-directory? file-is-directory?)
(define (file-regular? f)
  (eq? 'regular (stat:type (stat f))))

(define (make-file-permission-predicate user group world)
  (lambda (file)
    (let* ((info (stat file))
           (mode (stat:mode info)))
      (or (not (zero? (logand mode world)))
          (and (not (zero? (logand mode group)))
               (eqv? (stat:gid info) (passwd:gid (getpwnam (cuserid)))))
          (and (not (zero? (logand mode user)))
               (eqv? (stat:uid info) (passwd:uid (getpwnam (cuserid)))))))))

(define file-readable? (make-file-permission-predicate #o400 #o040 #o004))
(define file-writable? (make-file-permission-predicate #o200 #o020 #o002))

(define (file-size f)
  (stat:size (stat f)))

(define (file-modification-time f)
  (stat:mtime (stat f)))

(define (directory-fold dir kons knil)
  (let ((p (opendir dir)))
    (readdir p)
    (readdir p)
    (let lp ((acc knil))
      (let ((s (readdir p)))
        (if (string? s)
          (lp (kons s acc))
          (begin
            (closedir p)
            acc))))))

(define (directory-files dir)
  (directory-fold dir cons '()))

(define (directory-fold-tree dir kons knil)
  (let lp ((ls (directory-files dir))
           (d dir)
           (acc knil))
    (cond
      ((null? ls)
       acc)
      ((let ((d2 (string-append d "/" (car ls))))
         (and (file-directory? d2) d2))
       =>
       (lambda (d2)
         (lp (cdr ls)
             d
             (kons d (car ls)
                   (lp (directory-files d2) d2 acc)))))
      (else
       (lp (cdr ls) d (kons d (car ls) acc))))))

;; search-path doesn't allow a predicate
(define (find-in-path file ls . o)
  (let ((pred (if (pair? o) (car o) (lambda (f) #t))))
    (let lp ((ls ls))
      (and (pair? ls)
           (let* ((dir (car ls))
                  (len (string-length dir))
                  (path (if (eqv? #\/ (string-ref dir (- len 1)))
                          (string-append dir file)
                          (string-append dir "/" file))))
             (if (and (file-exists? path) (pred path))
               path
               (lp (cdr ls))))))))

(define (create-directory* f)
  (let* ((end (- (string-length f) 1))
         (f/ (if (eqv? #\/ (string-ref f end)) (substring f 0 end) f)))
    (or (file-exists? f/)
        (catch #t (lambda () (create-directory f/))
         (lambda e
           (let ((dir (dirname f/)))
             (and (string? dir)
                  (not (string=? dir f/))
                  (create-directory* dir)
                  (create-directory f/))))))))

(define delete-file
  (let ((unlink delete-file))
    (lambda (f) (catch #t (lambda () (unlink f)) (lambda e (rmdir f))))))
