
(define-module (common process))

(use-modules (ice-9 popen) (ice-9 rdelim) (srfi srfi-13))

(export process-run process-run-and-exit
        process-output->string process-output->string-list
        with-input-from-process with-output-to-process with-process-io
        shell-escape)

(define (shell-escape-char? c)
  (not (or (char-alphabetic? c)
           (char-numeric? c)
           (memv c '(#\_ #\- #\+ #\: #\, #\= #\/ #\.)))))

(define (shell-escape-backslash str)
  (with-output-to-string
    (lambda ()
      (with-input-from-string str
        (lambda ()
          (let lp ()
            (let ((c (read-char)))
              (if (not (eof-object? c))
                (begin
                  (if (shell-escape-char? c)
                    (display #\\))
                  (display c)
                  (lp))))))))))

(define (shell-escape str)
  (let lp ((i (- (string-length str) 1)) (quote? #f))
    (if (negative? i)
      (if quote?
        (string-append "\"" str "\"")
        str)
      (let ((c (string-ref str i)))
        (cond
          ((or (eqv? c #\") (eqv? c #\\))
           (shell-escape-backslash str))
          ((shell-escape-char? c)
           (lp (- i 1) #t))
          (else
           (lp (- i 1) quote?)))))))

(define (make-process-string command ls)
  (define (escape x)
    (shell-escape (if (string? x) x (object->string x))))
  (string-join (cons command (map escape ls))))

(define (process-run x . args)
  (zero? (or (status:exit-val (system (make-process-string x args))) 1)))

(define (process-run-and-exit x . args)
  (exit (or (status:exit-val (system (make-process-string x args))) 1)))

(define (with-input-from-process x thunk)
  (let* ((p (open-input-pipe
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (res (with-input-from-port p thunk)))
    (close-input-port p)
    res))

(define (with-output-to-process x thunk)
  (let* ((p (open-output-pipe
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (res (with-output-to-port p thunk)))
    (close-output-port p)
    res))

(define (with-process-io x thunk)
  (let* ((p (open-input-output-pipe
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (res (with-input-from-port p
                (lambda () (with-output-to-port p thunk)))))
    (close-port p)
    (res)))

(define (process-output->string-list x . args)
  (let ((command (make-process-string x args)))
    (with-input-from-process command
      (lambda ()
        (let lp ((res '()))
          (let ((line (read-line)))
            (if (eof-object? line)
              (reverse res)
              (lp (cons line res)))))))))

(define (process-output->string . o)
  (string-join (apply process-output->string-list o) "\n"))

