
(module common/process
  (process-run process-run-and-exit shell-escape
   process-output->string process-output->string-list
   with-input-from-process with-output-to-process with-process-io)

(import os)

(define (shell-escape-char? c)
  (not (or (char-alphabetic? c)
           (char-numeric? c)
           (memv c '(#\_ #\- #\+ #\: #\, #\= #\/ #\.)))))

(define (shell-escape-backslash str)
  (with-output-to-string
    (lambda ()
      (with-input-from-string str
        (lambda ()
          (let lp ()
            (let ((c (read-char)))
              (if (not (eof-object? c))
                (begin
                  (if (shell-escape-char? c)
                    (display #\\))
                  (display c)
                  (lp))))))))))

(define (shell-escape str)
  (let lp ((i (- (string-length str) 1)) (quote? #f))
    (if (negative? i)
      (if quote?
        (string-append "\"" str "\"")
        str)
      (let ((c (string-ref str i)))
        (cond
          ((or (eqv? c #\") (eqv? c #\\))
           (shell-escape-backslash str))
          ((shell-escape-char? c)
           (lp (- i 1) #t))
          (else
           (lp (- i 1) quote?)))))))

(define (display-to-string x)
  (cond
    ((symbol? x) (symbol->string x))
    ((number? x) (number->string x))
    (else (with-output-to-string (lambda () (display x))))))

(define (make-process-string command ls)
  (define (escape x)
    (shell-escape (if (string? x) x (x->string x))))
  (string-join (cons command (map escape ls))))

(define (process-run x . args)
  (let ((res (wait-for-process
              (spawn-process (make-process-string x args)))))
    (and (number? res) (zero? res))))

(define (process-run-and-exit x . args)
  (exit (or (wait-for-process
             (spawn-process (make-process-string x args)))
            1)))

(define (with-input-from-process x thunk)
  (let* ((p (spawn-process
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (in (get-process-stdin p)))
    (with-input-from-port in thunk)))

(define (with-output-to-process x thunk)
  (let* ((p (spawn-process
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (out (get-process-stdout p)))
    (with-output-to-port out thunk)))

(define (with-process-io x thunk)
  (let* ((p (spawn-process
             (if (pair? x) (make-process-string (car x) (cdr x)) x)))
         (in (get-process-stdin p))
         (out (get-process-stdout p)))
    (with-input-from-port in
      (lambda () (with-output-to-port out thunk)))))

(define (process-output->string-list x . args)
  (let ((command (make-process-string x args)))
    (with-input-from-process command
      (lambda ()
        (let lp ((res '()))
          (let ((line (read-line)))
            (if (eof-object? line)
              (reverse res)
              (lp (cons line res)))))))))

(define (process-output->string . o)
  (string-intersperse (apply process-output->string-list o) "\n"))

)
