
(require-extension (srfi 10000))

(common-module ()
 ((import-syntax (srfi testing))
  (import-extension (common (env pathname)))
  (entry-point main))

(define (main args)

(test-begin "pathname")

;; tests from the dirname(3) manpage

(test-equal "dirname(3)" "/usr" (path-directory "/usr/lib"))
(test-equal "lib"  (path-strip-directory "/usr/lib"))

(test-equal "/"    (path-directory "/usr/"))
(test-equal ""     (path-strip-directory "/usr/"))

(test-equal "."    (path-directory "usr"))
(test-equal "usr"  (path-strip-directory "usr"))

(test-equal "/"    (path-directory "/"))
(test-equal ""     (path-strip-directory "/"))

(test-equal "."    (path-directory "."))
(test-equal "."    (path-strip-directory "."))

(test-equal "."    (path-directory ".."))
(test-equal ".."   (path-strip-directory ".."))

;; additional tests (should match GNU dirname/basename behavior)

(test-equal "path-directory:border"
            "/"    (path-directory "//"))
(test-equal ""     (path-strip-directory "//"))

(test-equal "."    (path-directory ""))
(test-equal ""     (path-strip-directory ""))

(test-equal "."    (path-directory "../"))
(test-equal ""     (path-strip-directory "../"))

(test-equal ".."   (path-directory "../.."))
(test-equal ".."   (path-strip-directory "../.."))

(test-equal "path-directory:extra"
            "/usr/local" (path-directory "/usr/local/lib"))
(test-equal "lib" (path-strip-directory "/usr/local/lib"))

(test-equal "/usr" (path-directory "/usr/local/"))
(test-equal "" (path-strip-directory "/usr/local/"))

(test-equal "usr"    (path-directory "usr/local"))
(test-equal "local"  (path-strip-directory "usr/local"))

(test-equal "/" (path-directory "//usr"))
(test-equal "usr" (path-strip-directory "//usr"))

(test-equal "/" (path-directory "//usr/"))
(test-equal "" (path-strip-directory "//usr/"))

(test-equal "path-directory:small"
            "/a" (path-directory "/a/b"))
(test-equal "b" (path-strip-directory "/a/b"))

(test-equal "a" (path-directory "a/b"))
(test-equal "b" (path-strip-directory "a/b"))

(test-equal "a" (path-directory "a/b/"))
(test-equal "" (path-strip-directory "a/b/"))

(test-equal "/a/b/c" (path-directory "/a/b/c/d"))
(test-equal "d" (path-strip-directory "/a/b/c/d"))

(test-equal "/a/b/c" (path-directory "/a/b/c/d/"))
(test-equal "" (path-strip-directory "/a/b/c/d/"))

(test-equal "a/b/c"    (path-directory "a/b/c/d"))
(test-equal "d"  (path-strip-directory "a/b/c/d"))

(test-equal "/a/b" (path-directory "/a/b/c.d"))
(test-equal "c.d" (path-strip-directory "/a/b/c.d"))

(test-equal "/a/b" (path-directory "/a/b/c.d/"))
(test-equal "" (path-strip-directory "/a/b/c.d/"))

(test-equal "/a/b/c" (path-directory "/a/b/c/."))
(test-equal "." (path-strip-directory "/a/b/c/."))

(test-equal "/a/b/c" (path-directory "/a/b/c/.."))
(test-equal ".." (path-strip-directory "/a/b/c/.."))

(test-equal "/a/b/." (path-directory "/a/b/./c"))
(test-equal "c" (path-strip-directory "/a/b/./c"))

(test-equal "/a/b/.." (path-directory "/a/b/../c"))
(test-equal "c" (path-strip-directory "/a/b/../c"))

(test-equal "/a/b" (path-directory "/a/b/c//"))
(test-equal "" (path-strip-directory "/a/b/c//"))

(test-equal "/a/b" (path-directory "/a/b//c///"))
(test-equal "" (path-strip-directory "/a/b//c///"))

;; extensions

(test-equal "path-extension" "scm" (path-extension "foo.scm"))
(test-equal "foo" (path-strip-extension "foo.scm"))

(test-equal "c" (path-extension "foo.scm.c"))
(test-equal "foo.scm" (path-strip-extension "foo.scm.c"))

(test-equal "scm" (path-extension "/home/me/foo.scm"))
(test-equal "/home/me/foo" (path-strip-extension "/home/me/foo.scm"))

(test-equal "scm" (path-extension "foo..scm"))
(test-equal "foo." (path-strip-extension "foo..scm"))

(test-equal "s" (path-extension "foo.s"))
(test-equal "foo" (path-strip-extension "foo.s"))

(test-equal #f (path-extension "foo."))
(test-equal "foo." (path-strip-extension "foo."))

(test-equal #f (path-extension "foo.scm."))
(test-equal "foo.scm." (path-strip-extension "foo.scm."))

(test-equal #f (path-extension "."))
(test-equal "." (path-strip-extension "."))

(test-equal #f (path-extension "a."))
(test-equal "a." (path-strip-extension "a."))

(test-equal #f (path-extension "/."))
(test-equal "/." (path-strip-extension "/."))

(test-equal #f (path-extension "foo.scm/"))
(test-equal "foo.scm/" (path-strip-extension "foo.scm/"))

(test-equal "path-replace-extension"
            "foo.c" (path-replace-extension "foo.scm" "c"))
(test-equal "foo.c" (path-replace-extension "foo" "c"))

;; absolute paths

(test-assert (path-absolute? "/"))
(test-assert (path-absolute? "//"))
(test-assert (path-absolute? "/usr"))
(test-assert (path-absolute? "/usr/"))
(test-assert (path-absolute? "/usr/."))
(test-assert (path-absolute? "/usr/.."))
(test-assert (path-absolute? "/usr/./"))
(test-assert (path-absolute? "/usr/../"))

(test-assert (not (path-absolute? "")))
(test-assert (not (path-absolute? ".")))
(test-assert (not (path-absolute? "usr")))
(test-assert (not (path-absolute? "usr/")))

;; normalization & building

(test-equal "path-normalize" "/a/b/c/d/e" (path-normalize "/a/b/c/d/./e"))
(test-equal "/a/b/c/d/e" (path-normalize "/a/b//.///c//d/./e"))
(test-equal "/a/b/c/d/e/" (path-normalize "/a/b//.///c//d/./e/"))
(test-equal "/a/c/d/e" (path-normalize "/a/b/../c/d/e"))
(test-equal "/a/b/c/e" (path-normalize "/a/b//.///c//d/../e"))
(test-equal "/a/c/e" (path-normalize "/a/b//..///c//d/../e"))
(test-equal "/a/b/c/d/e/"
            (path-normalize "/a/b//./../c/d/../../b//c/d/e/f/.."))

(test-equal "path-normalize:border" "" (path-normalize ""))
(test-equal "." (path-normalize "."))
(test-equal "/" (path-normalize "/"))

(test-equal "path-normalize:overflow"
            "/" (path-normalize "/a/b/c/../../../../.."))
(test-equal "a/../../.." (path-normalize "a/b/c/../../../../.."))
(test-equal "../../.." (path-normalize "../a/b/c/../../../../.."))

(test-equal "path-expand" "/a" (path-expand "" "/a"))
(test-equal "/a/b" (path-expand "b" "/a"))
(test-equal "/b" (path-expand "/b" "/a"))
(test-equal (string-append (getenv "HOME") "/a") (path-expand "~/a"))
(test-equal (string-append (getenv "HOME") "/a")
            (path-expand
             (string-append "~" (or (getenv "USER") (getenv "LOGIN")) "/a")))

(test-equal "path-relative" "c" (path-relative "/a/b/c" "/a/b"))
(test-equal #f (path-relative "/d/a/b/c" "/a/b"))

(test-equal "path-build" "a/b" (path-build "a" "b"))
(test-equal "a/b" (path-build "a/" "b"))
(test-equal "a/b/./c" (path-build "a" "b" "." "c"))
(test-equal "a/b/../c" (path-build "a" "b" ".." "c"))
(test-equal "a/b/c" (path-build "a" '("b" "c")))

(test-end "pathname")

))
